#include "editorContext.h"

#include <algorithm>
#include <functional>
#include <vector>
#include <string>
#include <cstdlib>
#include <cctype>

#include <unknwn.h>
#include <objidl.h>
#include <objbase.h>

#include "helpContexts.h"
#include "farManager.h"
#include "farEditor.h"
#include "parsing.h"
#include "findWindow.h"
#include "shiftDialog.h"
#include "ssaTime.h"
#include "exception.h"
#include "rot.h"
#include "oleString.h"
#include "language.h"

// these IDs are subject to change in future versions of Media Player Classic
// from resource.h:
const WPARAM ID_PLAY_FRAMESTEP = 891;
const WPARAM ID_PLAY_FRAMESTEPCANCEL = 892;
const WPARAM ID_SUBTITLES_SUBITEM_START = 2300;
// from code in MainFrm.cpp, function CMainFrame::SetupSubtitlesSubMenu:
const WPARAM ID_SUBTITLES_RELOAD = ID_SUBTITLES_SUBITEM_START + 2;

const char* PHRASE_TEMPLATE = "Dialogue: Marked=0,0:00:00.00,0:00:00.00,*Default,,0000,0000,0000,,";

LONGLONG inputTime(LONGLONG time, const char* title = 0, const char* prompt = 0,
	const char* historyName = 0, const char* helpTopic = 0)
{
	char s[11];
	formatTime(time, s, s + 10);
	s[10] = 0;
	if (!farManager.inputBox(title, prompt, historyName, s, s, 11, helpTopic, 0))
		return time;
	LONGLONG result = parseTime(s, s + 10);
	if (INVALID_TIME == result)
		return time;
	return result;
}

EditorContext::EditorContext()
:	pid_(0), hwndPlayer_(0)
{
}

bool EditorContext::attached() const
{
	return pid_ != 0;
}

bool EditorContext::paused() const
{
	if (!attached()) return true;

	OAFilterState filterState;
	ASSERT_SUCCESS(pMediaControl_->GetState(1500, &filterState), "Cannot get player state");

	return filterState == State_Stopped || filterState == State_Paused;
}

void EditorContext::run()
{
	if (!attached()) return;

	ASSERT_SUCCESS(pMediaControl_->Run(), "Cannot start playback");
}

void EditorContext::pause()
{
	if (!attached()) return;

	ASSERT_SUCCESS(pMediaControl_->Pause(), "Cannot pause playback");
}

LONGLONG EditorContext::mediaTime() const
{
	if (!attached()) return INVALID_TIME;

	LONGLONG position;
	ASSERT_SUCCESS(pMediaSeeking_->GetCurrentPosition(&position), "Cannot get media time");
	return position;
}

void EditorContext::seek(LONGLONG time)
{
	if (!attached()) return;

	ASSERT_SUCCESS(pMediaSeeking_->SetPositions(
		&time, AM_SEEKING_AbsolutePositioning,
		&time, AM_SEEKING_NoPositioning), "Seek error");
}

void EditorContext::previousFrame()
{
	if (!attached()) return;

	PostMessage(hwndPlayer_, WM_COMMAND, ID_PLAY_FRAMESTEPCANCEL, 0);
}

void EditorContext::nextFrame()
{
	if (!attached()) return;

	PostMessage(hwndPlayer_, WM_COMMAND, ID_PLAY_FRAMESTEP, 0);
}

template <typename InIter>
void
EditorContext::setDialogueField(size_t fieldNo, InIter valueBegin, InIter valueEnd)
{
	EditorLine line = farEditor.currentLine();

	if (!starts_with(line, "Dialogue:")) return;

	const char
		*const begin	= comma(fieldNo, line.begin(), line.end()),
		*const end		= next_comma(begin, line.end());
	if (begin == line.end() || end == line.end()) return; // syntax

	const size_t oldSize = (end - (begin + 1));
	std::vector<char> newLine(line.size() - oldSize + std::distance(valueBegin, valueEnd));

	std::copy			(end,			line.end(),
		std::copy		(valueBegin,	valueEnd,
			std::copy	(line.begin(),	begin + 1,
				newLine.begin())));

	line.set(&newLine[0], newLine.size());
	farEditor.invalidate();
}

void EditorContext::setStyle(const std::string& styleName)
{
	// FIXME: assumes default Dialogue format
	setDialogueField(2, styleName.begin(), styleName.end());
}

void EditorContext::setStartTime(LONGLONG time)
{
	if (time < 0) return;
	// FIXME: assumes default Dialogue format
	char value[11];
	formatTime(time, value, value + 10);
	setDialogueField(0, value, value + 10);
}

void EditorContext::setEndTime(LONGLONG time)
{
	if (time < 0) return;
	// FIXME: assumes default Dialogue format
	char value[11];
	formatTime(time, value, value + 10);
	setDialogueField(1, value, value + 10);
}

template <typename InIter>
void EditorContext::setText(InIter begin, InIter end)
{
	EditorLine line = farEditor.currentLine();

	if (!starts_with(line, "Dialogue:")) return;

	EditorLine::const_iterator headerEnd = comma(8, line.begin(), line.end());
	if (headerEnd == line.end()) return;

	std::vector<char> newText(line.begin(), headerEnd + 1);
	newText.insert(newText.end(), begin, end);

	line.set(&newText[0], newText.size());
	farEditor.invalidate();
}

bool EditorContext::scrollLock_;

void EditorContext::updateScrollLock(const INPUT_RECORD& rec)
{
	switch (rec.EventType)
	{
	case KEY_EVENT:
	case FARMACRO_KEY_EVENT:
		// FAR has an undocumented feature
		// where a null key event is fired each second.
		if (0 == rec.Event.KeyEvent.wVirtualKeyCode &&
			0 == rec.Event.KeyEvent.wVirtualScanCode &&
			0 == rec.Event.KeyEvent.uChar.AsciiChar) break;
		scrollLock_ = rec.Event.KeyEvent.dwControlKeyState & SCROLLLOCK_ON;
		break;
	case MOUSE_EVENT:
		scrollLock_ = rec.Event.MouseEvent.dwControlKeyState & SCROLLLOCK_ON;
		break;
	}
}

bool EditorContext::tracking()
{
	return attached() && !paused() && scrollLock_;
}

void EditorContext::togglePause()
{
	if (!attached()) return;

	if (paused())
	{
		run();
	}
	else
	{
		pause();
	}
}

class Pause
{
public:
	Pause(EditorContext& editor)
	: editor_(editor), wasPaused_(editor.paused())
	{
		editor.pause();
	}
	~Pause()
	{
		if (!wasPaused_)
		{
			editor_.run();
		}
	}
private:
	EditorContext& editor_;
	bool wasPaused_;
};

void EditorContext::editPhrase()
{
	Pause p(*this);

	EditorLine line = farEditor.currentLine();
	if (!starts_with(line, "Dialogue:")) return;

	const char* begin = comma(8, line.begin(), line.end());
	if (begin == line.end()) return; // syntax
	++begin;

	std::vector<char> phrase(begin, line.end());

	farEditor.convertToOem(&phrase[0], phrase.size());

	std::vector<FarDialogItem> items;

	std::vector<char>::const_iterator i = phrase.begin(), end = phrase.end();

	const char newLine[] = "\\N";
	while (i != end || items.size() < 3)
	{
		std::vector<char>::const_iterator sep = std::search(i, end, newLine, newLine + 2);
		items.resize(items.size() + 1);
		items.back().Type = DI_EDIT;
		items.back().X1 = 3; items.back().Y1 = items.size(); items.back().X2 = 42; items.back().Y2 = 0;
		items.back().Focus = 1 == items.size();
		items.back().History = 0;
		items.back().Flags = DIF_EDITOR;
		items.back().DefaultButton = 1 == items.size();
		std::copy(i, i + (std::min)(512, sep - i), items.back().Data);

		i = sep + (sep == end ? 0 : 2);
	}

	if (-1 == farManager.dialog(&items[0], items.size(), 46, 2 + items.size(), help::editPhrase)) return;

	phrase.assign(items[0].Data, std::find(
		items[0].Data, items[0].Data + COUNTOF(items[0].Data), 0));
	for (std::vector<FarDialogItem>::const_iterator i = items.begin() + 1, e = items.end();
		i != e; ++i)
	{
		if (i->Data[0] != 0)
		{
			phrase.insert(phrase.end(), newLine, newLine + 2);
			phrase.insert(phrase.end(), i->Data, std::find(
				i->Data, i->Data + COUNTOF(i->Data), 0));
		}
	}

	farEditor.convertFromOem(&phrase[0], phrase.size());

	phrase.insert(phrase.begin(), line.begin(), begin);
	line.set(&phrase[0], phrase.size());
}

void EditorContext::splitPhraseSimultaneous()
{
	EditorLine line = farEditor.currentLine();
	if (!starts_with(line, "Dialogue:")) return;

	LONGLONG
		startTime = getStartTime(line),
		endTime = getEndTime(line);
	std::string style = getStyle(line);

	EditorPositionSaver info;

	int phrasePos = std::distance(line.begin(), comma(8, line.begin(), line.end())) + 1;
	if (info.CurPos < phrasePos)
	{
		farEditor.setPosition(DEFAULT, phrasePos);
	}

	farEditor.insertLine();
	farEditor.insertText(PHRASE_TEMPLATE);

	setStartTime(startTime);
	setEndTime(endTime);
	setStyle(style);
}

void EditorContext::splitPhraseSequential()
{
	EditorLine line = farEditor.currentLine();
	if (!starts_with(line, "Dialogue:")) return;

	LONGLONG
		startTime = getStartTime(line),
		endTime = getEndTime(line),
		splitTime = mediaTime();
	if (splitTime < startTime || endTime <= splitTime)
	{
		splitTime = (startTime + endTime) / 2;
	}
	std::string style = getStyle(line);

	EditorPositionSaver info;

	setEndTime(splitTime);

	int phrasePos = std::distance(line.begin(), comma(8, line.begin(), line.end())) + 1;
	if (info.CurPos < phrasePos)
	{
		farEditor.setPosition(DEFAULT, phrasePos);
	}

	farEditor.insertLine();
	farEditor.insertText(PHRASE_TEMPLATE);

	setStartTime(splitTime);
	setEndTime(endTime);
	setStyle(style);
}

void EditorContext::splicePhrase()
{
	EditorPositionSaver info;

	EditorLine line = farEditor.currentLine();
	if (!starts_with(line, "Dialogue:")) return;

	farEditor.setPosition(info.CurLine + 1);
	EditorLine nextLine = farEditor.currentLine();
	if (!starts_with(line, "Dialogue:")) return;

	LONGLONG endTime = getEndTime(nextLine);

	EditorLine::const_iterator phraseBegin = comma(8, nextLine.begin(), nextLine.end());
	if (phraseBegin == nextLine.end()) return;

	++phraseBegin;
	const char newLine[] = "\\N";
	std::vector<char> newText;
	newText.reserve(line.size() + 2 + std::distance(phraseBegin, nextLine.end()));
	newText.insert(newText.end(), line.begin(), line.end());
	newText.insert(newText.end(), newLine,      newLine + 2);
	newText.insert(newText.end(), phraseBegin,  nextLine.end());

	farEditor.deleteLine();

	info.revert();

	line.set(&newText[0], newText.size());

	setEndTime(endTime);
}

class AddLine : public std::binary_function<
	EditorLine::const_iterator, EditorLine::const_iterator, void>
{
public:
	explicit AddLine(std::vector<std::string>& lines)
	: lines_(lines)
	{
	}

	void operator()(EditorLine::const_iterator begin, EditorLine::const_iterator end) const
	{
		lines_.push_back(std::string(begin, end));
	}
private:
	std::vector<std::string>& lines_;
};

void EditorContext::fixOverlay()
{
	typedef std::vector<char>::iterator iterator;

	EditorPositionSaver info;

	EditorLine currentL = farEditor.currentLine();
	if (!starts_with(currentL, "Dialogue:")) return;
	std::vector<char> current(currentL.begin(), currentL.end());
	LONGLONG currentStartTime = getStartTime(current), currentEndTime = getEndTime(current);
	std::string style = getStyle(current);
	iterator currentText = comma(8, current.begin(), current.end());
	if (currentText == current.end()) return;
	++currentText;

	farEditor.setPosition(info.CurLine + 1, DEFAULT, info.CurTabPos);

	EditorLine nextL = farEditor.currentLine();
	if (!starts_with(nextL, "Dialogue:")) return;
	std::vector<char> next(nextL.begin(), nextL.end());
	LONGLONG nextStartTime = getStartTime(next), nextEndTime = getEndTime(next);
	iterator nextText = comma(8, next.begin(), next.end());
	if (nextText == next.end()) return;
	++nextText;
	if (nextStartTime != currentEndTime) return;

	const char newLine[] = "\\N";
	for (iterator begin = currentText; begin != current.end();
		begin = afterNextSeparator(begin, current.end(), newLine, newLine + 2))
	{
		for (iterator end = current.end(); end != begin;
			end = previousSeparator(begin, end, newLine, newLine + 2))
		{
			std::iterator_traits<iterator>::difference_type
				overlayLength = std::distance(begin, end);
			for (iterator overlayBegin = nextText; overlayBegin != next.end();
				overlayBegin = afterNextSeparator(overlayBegin, next.end(), newLine, newLine + 2))
			{
				if (// too long
					overlayLength > std::distance(overlayBegin, next.end()) ||
					// not line end
					!(overlayBegin + overlayLength == next.end() ||
					overlayBegin[overlayLength] == newLine[0] &&
					overlayBegin + overlayLength + 1 != next.end() &&
					overlayBegin[overlayLength + 1] == newLine[1]) ||
					// not equal
					!std::equal(begin, end, overlayBegin))
				{
					continue;
				}

				// add a phrase containing the overlay
				farEditor.setPosition(info.CurLine + 1, DEFAULT, 0);
				farEditor.insertText(PHRASE_TEMPLATE);
				farEditor.insertLine();
				farEditor.setPosition(info.CurLine + 1, DEFAULT, 0);
				setStartTime(currentStartTime);
				setEndTime(nextEndTime);
				setStyle(style);
				setText(begin, end);

				farEditor.setPosition(info.CurLine + 2, DEFAULT, 0);
				bool hasSuffix = overlayBegin + overlayLength != next.end();
				if (nextText == overlayBegin && !hasSuffix)
				{
					farEditor.deleteLine();
				}
				else
				{
					if (hasSuffix)
					{
						next.erase(overlayBegin, overlayBegin + overlayLength + 2);
					}
					else
					{
						next.erase(overlayBegin - 2, overlayBegin + overlayLength);
					}
					nextL.set(&next[0], next.size());
				}

				info.revert();
				hasSuffix = end != current.end();
				if (currentText == begin && !hasSuffix)
				{
					farEditor.deleteLine();
				}
				else
				{
					if (hasSuffix)
					{
						current.erase(begin, end + 2);
					}
					else
					{
						current.erase(begin - 2, end);
					}
					currentL.set(&current[0], current.size());
				}

				farEditor.invalidate();
				return;
			}
		}
	}
}

void EditorContext::movePhraseUp()
{
	GetEditorInfo info;
	if (info.CurLine == 0) return;

	EditorLine line = farEditor.currentLine();
	std::vector<char> s(line.begin(), line.end());

	farEditor.deleteLine();

	farEditor.setPosition(info.CurLine - 1, 0);
	farEditor.insertLine();

	farEditor.setPosition(info.CurLine - 1, 0);
	farEditor.currentLine().set(&s[0], s.size());

	farEditor.setPosition(info.CurLine - 1, DEFAULT, info.CurTabPos,
		(std::max)(0, info.TopScreenLine - 1), info.LeftPos);
}

void EditorContext::movePhraseDown()
{
	GetEditorInfo info;
	if (info.CurLine + 1 == info.TotalLines) return;

	EditorLine line = farEditor.currentLine();
	std::vector<char> s(line.begin(), line.end());

	farEditor.deleteLine();

	farEditor.setPosition(info.CurLine + 1, 0);
	farEditor.insertLine();

	farEditor.setPosition(info.CurLine + 1, 0);
	farEditor.currentLine().set(&s[0], s.size());

	farEditor.setPosition(info.CurLine + 1, DEFAULT, info.CurTabPos,
		info.TopScreenLine + 1, info.LeftPos);
}

struct MediaPlayer
{
	DWORD pid;
	HWND hwnd;
	CComPtr<IUnknown> pFilterGraph;

	MediaPlayer(DWORD aPid = 0,
		HWND anHwnd = 0,
		CComPtr<IUnknown> apFilterGraph = 0)
	: pid(aPid), hwnd(anHwnd), pFilterGraph(apFilterGraph)
	{
	}
};

class GetName
{
public:
	explicit GetName(HWND hwnd)
	: hwnd_(hwnd)
	{
	}

	void operator()(FarMenuItem& item) const
	{
		SendMessage(hwnd_, WM_GETTEXT,
			COUNTOF(item.Text), reinterpret_cast<LPARAM>(item.Text));
	}
private:
	HWND hwnd_;
};

bool EditorContext::attachMenu()
{
	FarMenuItems items;
	std::vector<MediaPlayer> players(2);

	items
		(lng::attach::detach)(SelectIf(!attached()))
		();

	RunningObjectTable ROT;

	CComPtr<IBindCtx> pBindCtx;
	ASSERT_SUCCESS(CreateBindCtx(0, &pBindCtx), "Cannot create bind context");

	CComPtr<IEnumMoniker> pEnum = ROT.EnumRunning();
	for (CComPtr<IMoniker> pMoniker; S_OK == pEnum->Next(1, &pMoniker, 0); pMoniker = 0)
	{
		OleString displayName;
		ASSERT_SUCCESS(pMoniker->GetDisplayName(pBindCtx, 0, &displayName), "Cannot get display name");

		const wchar_t mediaPlayerClassicToken[] = L"(MPC)";
		if (displayName.end() == std::search(displayName.begin(), displayName.end(),
			mediaPlayerClassicToken, mediaPlayerClassicToken + COUNTOF(mediaPlayerClassicToken) - 1))
			continue;

		const wchar_t pidToken[] = L" pid ";
		const wchar_t* pidTokenBegin =
			std::search(displayName.begin(), displayName.end(), pidToken, pidToken + COUNTOF(pidToken) - 1);
		if (displayName.end() == pidTokenBegin) continue;
		pidTokenBegin += COUNTOF(pidToken) - 1; // skip keyword

		wchar_t* pidTokenEnd;
		DWORD pid = std::wcstol(pidTokenBegin, &pidTokenEnd, 16);
		if (pidTokenEnd != pidTokenBegin + 8) continue;

		FindMpcWindowByPid fw(pid);
		if (!fw.run()) continue;

		items("")(SelectIf(pid_ == pid))(GetName(fw.hwnd()));
		players.push_back(MediaPlayer(pid, fw.hwnd(), ROT.GetObject(pMoniker)));
	}

	int result = farManager.menu(items, lng::attach::title, 0, help::attach);
	if (result < 0) return false;

	pid_ = players[result].pid;
	hwndPlayer_ = players[result].hwnd;
	pMediaSeeking_ = players[result].pFilterGraph;
	pMediaControl_ = players[result].pFilterGraph;
	return true;
}

bool EditorContext::seekMenu()
{
	switch (farManager.menu(FarMenuItems()
		(lng::seek::start        )
		(lng::seek::end          )
		(                        )
		(lng::seek::previousFrame)
		(lng::seek::nextFrame    )
		(                        )
		(lng::seek::custom       ),
		lng::seek::title, 0, help::seek))
	{
	case 0:
		seek(getStartTime(farEditor.currentLine()));
		break;
	case 1:
		seek(getEndTime(farEditor.currentLine()));
		break;
	case 3:
		previousFrame();
		break;
	case 4:
		nextFrame();
		break;
	case 6:
		seek(inputTime(mediaTime(),
			lng::seek::dialog::title,
			lng::seek::dialog::prompt,
			"SSAEditor.Seek",
			("<" + std::string(farManager.moduleName()) + ">SeekCustom").c_str()));
		break;
	default:
		return false;
	}
	return true;
}

bool EditorContext::phraseMenu()
{
	switch (farManager.menu(FarMenuItems()
		(lng::phrase::edit             )
		(lng::phrase::splitSequential  )
		(lng::phrase::splitSimultaneous)
		(lng::phrase::splice           )
		(lng::phrase::fixOverlay       )
		(lng::phrase::moveUp           )
		(lng::phrase::moveDown         ),
		lng::phrase::title, 0, help::phrase))
	{
	case 0:
		editPhrase();
		break;
	case 1:
		splitPhraseSequential();
		break;
	case 2:
		splitPhraseSimultaneous();
		break;
	case 3:
		splicePhrase();
		break;
	case 4:
		fixOverlay();
		break;
	case 5:
		movePhraseUp();
		break;
	case 6:
		movePhraseDown();
		break;
	default:
		return false;
	}
	return true;
}

class ParseStyle : public std::unary_function<EditorLine, void>
{
private:
	FarMenuItems& items_;
	std::vector<std::string>& names_;
	std::string style_;
public:
	explicit ParseStyle(FarMenuItems& items,
		std::vector<std::string>& names, const std::string& style)
	: items_(items), names_(names), style_(style)
	{
	}

	void operator()(const EditorLine& line) const
	{
		if (!starts_with(line, "Style:")) return;
		const char
			*const begin	= std::find_if(line.begin() + 6, line.end(),
											std::not1(std::ptr_fun(std::isspace))),
			*const end		= next_comma(begin, line.end());
		if (begin == line.end() || end == line.end()) return;

		std::string name(begin, end);
		if (name == "Default")
		{
			name.insert(name.begin(), '*');
		}

		names_.push_back(name);
		items_("")(SelectIf(name == style_));
		char* out = items_.back().Text;
		if (items_.size() < 10)
		{
			*out++ = '&';
			*out++ = '0' + items_.size();
		}
		*out++ = ' ';
		std::copy(name.begin(), name.begin() + (std::min)
			(name.size(), COUNTOF(items_.back().Text) - (out - items_.back().Text)),
			out);
	}
};

bool EditorContext::styleMenu()
{
	FarMenuItems items;
	std::vector<std::string> names;

	farEditor.forEachLine(0, END_LINE,
		ParseStyle(items, names, getStyle(farEditor.currentLine())));

	int result = farManager.menu(items, lng::style::title, 0, help::style);
	if (-1 == result) return false;

	setStyle(names[result]);

	return true;
}

bool EditorContext::timingMenu()
{
	while (true)
	{
		switch (farManager.menu(FarMenuItemsEx()
			(lng::timing::setStart)(EnableIf(attached()))
			(lng::timing::setEnd  )(EnableIf(attached()))
			(                     )
			(lng::timing::shift   )
			(lng::timing::duration),
			lng::timing::title, 0, help::timing))
		{
		case 0:
			setStartTime(mediaTime());
			break;
		case 1:
			setEndTime(mediaTime());
			break;
		case 3:
			if (!shiftMenu()) continue;
			break;
		case 4:
			changeDuration();
			break;
		default:
			return false;
		}
		break;
	}
	return true;
}

void EditorContext::changeDuration()
{
	EditorPositionSaver info;
	EditorLine line = farEditor.currentLine();
	LONGLONG startTime = getStartTime(line), endTime = getEndTime(line);
	if (startTime < 0 || endTime < 0) return;

	ShiftDialog dlg(lng::timing::dialog::changeBy,
		lng::timing::dialog::endTime, endTime,
		lng::timing::dialog::duration, endTime - startTime,
		mediaTime(),
		help::duration);
	if (!dlg.execute()) return; // Cancelled dialog

	setEndTime(endTime + dlg.shift());
}

class InBlockAndHasValidTimes : public std::unary_function<EditorLine, bool>
{
private:
	LONGLONG& startTime_;
	LONGLONG& endTime_;
public:
	InBlockAndHasValidTimes(LONGLONG& startTime, LONGLONG& endTime)
	: startTime_(startTime), endTime_(endTime)
	{
	}
	bool operator()(const EditorLine& line) const
	{
		startTime_ = getStartTime(line);
		endTime_ = getEndTime(line);
		return line.selected() && startTime_ >= 0 && endTime_ >= 0;
	}
};

class HasValidTimes : public std::unary_function<EditorLine, bool>
{
private:
	LONGLONG& startTime_;
	LONGLONG& endTime_;
public:
	HasValidTimes(LONGLONG& startTime, LONGLONG& endTime)
	: startTime_(startTime), endTime_(endTime)
	{
	}
	bool operator()(const EditorLine& line) const
	{
		startTime_ = getStartTime(line);
		endTime_ = getEndTime(line);
		return startTime_ >= 0 && endTime_ >= 0;
	}
};

class ShiftPhrase : public std::unary_function<EditorLine, bool>
{
public:
	ShiftPhrase(EditorContext& editor, LONGLONG shift, bool blockOnly)
	: editor_(editor), shift_(shift), blockOnly_(blockOnly)
	{
	}

	bool operator()(EditorLine& line) const
	{
		if (blockOnly_ && !line.selected()) return true; // block has ended
		if (!starts_with(line, "Dialogue:")) return false; // not a phrase
		LONGLONG startTime = getStartTime(line), endTime = getEndTime(line);
		if (startTime >= 0)
		{
			editor_.setStartTime(startTime + shift_ >= 0 ? startTime + shift_ : 0);
		}
		if (endTime >= 0)
		{
			editor_.setEndTime(endTime + shift_ >= 0 ? endTime + shift_ : 0);
		}
		return false;
	}
private:
	EditorContext& editor_;
	LONGLONG shift_;
	bool blockOnly_;
};

bool EditorContext::shiftMenu()
{
	EditorPositionSaver info;

	EditorLine line = farEditor.currentLine();
	LONGLONG startTime = getStartTime(line), endTime = getEndTime(line);

	int scope = farManager.menu(FarMenuItemsEx()
		(lng::timing::scope::currentPhrase)(EnableIf(startTime >= 0 && endTime >= 0))
		(lng::timing::scope::selection)(EnableIf(info.BlockType != BTYPE_NONE))
		(lng::timing::scope::remainder)
		(lng::timing::scope::wholeScript),
		lng::timing::scope::title, 0, help::shift);

	switch (scope)
	{
	case 0:
		break;
	case 1:
		farEditor.findLineIf(info.BlockStartLine, END_LINE, InBlockAndHasValidTimes(startTime, endTime));
		break;
	case 2:
		farEditor.findLineIf(CURRENT_LINE, END_LINE, HasValidTimes(startTime, endTime));
		break;
	case 3:
		farEditor.findLineIf(0, END_LINE, HasValidTimes(startTime, endTime));
		break;
	default:
		return false;
	}
	if (startTime < 0 || endTime < 0) return true; // no lines to be processed, and we only found it out just now

	info.revert();

	ShiftDialog dlg(lng::timing::dialog::shiftBy,
		lng::timing::dialog::startTime, startTime,
		lng::timing::dialog::endTime, endTime,
		mediaTime(),
		help::shift);
	if (!dlg.execute()) return true; // Cancelled dialog

	ShiftPhrase shift(*this, dlg.shift(), scope == 1);

	switch (scope)
	{
	case 0:
		shift(line);
		break;
	case 1:
		farEditor.findLineIf(info.BlockStartLine, END_LINE, shift);
		break;
	case 2:
		farEditor.findLineIf(CURRENT_LINE, END_LINE, shift);
		break;
	case 3:
		farEditor.findLineIf(0, END_LINE, shift);
		break;
	}

	return true;
}

bool EditorContext::mainMenu()
{
	do
	{
		switch (farManager.menu(FarMenuItemsEx()
			(lng::main::attach)(CheckIf(attached()))
			(                 )
			(lng::main::pause )(EnableIf(attached()))(CheckIf(paused()))
			(lng::main::seek  )(EnableIf(attached()))
			(                 )
			(lng::main::phrase)
			(lng::main::style )
			(lng::main::timing),
			lng::main::title, 0, help::mainMenu))
		{
		case 0:
			if (!attachMenu()) continue;
			break;
		case 2:
			togglePause();
			break;
		case 3:
			if (!seekMenu()) continue;
			break;
		case 5:
			if (!phraseMenu()) continue;
			break;
		case 6:
			if (!styleMenu()) continue;
			break;
		case 7:
			if (!timingMenu()) continue;
			break;
		default:
			return false;
		}
		break;
	} while (true);
	return true;
}

bool EditorContext::isTrackingEvent(const INPUT_RECORD& rec) const
{
	return rec.EventType == KEY_EVENT &&
		rec.Event.KeyEvent.wVirtualKeyCode == 0;
}

class ContainsTime : public std::unary_function<EditorLine, bool>
{
public:
	explicit ContainsTime(LONGLONG time)
	: time_(time)
	{
	}

	bool operator()(const EditorLine& line) const
	{
		if (!starts_with(line, "Dialogue:")) return false;
		LONGLONG startTime = getStartTime(line), endTime = getEndTime(line);
		if (INVALID_TIME == startTime || INVALID_TIME == endTime) return false; // syntax
		return startTime <= time_ && time_ < endTime;
	}
private:
	LONGLONG time_;
};

class TimeIsAfter : public std::unary_function<EditorLine, bool>
{
public:
	explicit TimeIsAfter(LONGLONG time)
	: time_(time)
	{
	}

	bool operator()(const EditorLine& line) const
	{
		if (!starts_with(line, "Dialogue:")) return false;
		LONGLONG startTime = getStartTime(line), endTime = getEndTime(line);
		if (INVALID_TIME == startTime || INVALID_TIME == endTime) return false; // syntax
		return time_ < startTime;
	}
private:
	LONGLONG time_;
};

void EditorContext::track()
{
	if (!farEditor.findLineIf(0, END_LINE, ContainsTime(mediaTime())))
		farEditor.findLineIf(0, END_LINE, TimeIsAfter(mediaTime()));
	farEditor.redraw();
}

void EditorContext::saveAndReload()
{
	farEditor.save();
	SendMessage(hwndPlayer_, WM_COMMAND, ID_SUBTITLES_RELOAD, 0);
}

bool EditorContext::onKeyDown(WORD virtualKeyCode, DWORD controlKeyState)
{
	if (virtualKeyCode == VK_F2 && controlKeyState == 0)
	{
		saveAndReload();
		return true;
	}
	return false;
}

bool EditorContext::onKeyUp(WORD virtualKeyCode, DWORD controlKeyState)
{
	return false;
}

int EditorContext::processEditorInput(const INPUT_RECORD& rec)
{
	static const DWORD SHIFT_MASK =
		LEFT_ALT_PRESSED   |
		LEFT_CTRL_PRESSED  |
		RIGHT_ALT_PRESSED  |
		RIGHT_CTRL_PRESSED |
		SHIFT_PRESSED;
	updateScrollLock(rec);
	if (tracking() && isTrackingEvent(rec))
	{
		track();
	}

	if (rec.EventType == KEY_EVENT || rec.EventType == FARMACRO_KEY_EVENT)
	{
		if (rec.Event.KeyEvent.bKeyDown)
		{
			return onKeyDown(rec.Event.KeyEvent.wVirtualKeyCode,
				rec.Event.KeyEvent.dwControlKeyState & SHIFT_MASK) ? 1 : 0;
		}
		else
		{
			return onKeyUp(rec.Event.KeyEvent.wVirtualKeyCode,
				rec.Event.KeyEvent.dwControlKeyState & SHIFT_MASK) ? 1 : 0;
		}
	}

	return 0;
}
